from dataclasses import dataclass, field
import json
import logging
import re
from typing import List, Optional, Tuple, Union
from xml.etree import ElementTree as ET

import dataclass_wizard
from requests import Session

from .models import TracksEndpointStreamResponseJSON
from .utils import decrypt_manifest_key_id

logger = logging.getLogger("__name__")


class TidalManifestException(Exception):
    pass


@dataclass
class S:
    d: str
    r: Optional[str] = field(default=None)

    def __post_init__(self):
        self.d: Optional[int] = int(self.d) if self.d is not None else None
        self.r: Optional[int] = int(self.r) if self.r is not None else None


@dataclass(frozen=True)
class SegmentTimeline:
    s: Tuple[Optional["S"]]


@dataclass
class JSONDASHManifest:
    mime_type: Optional[str] = field(default=None)
    codecs: Optional[str] = field(default=None)
    encryption_type: Optional[str] = field(default=None)
    key_id: Optional[str] = field(default=None)
    urls: Optional[List[str]] = field(repr=False, default=None)

    def __post_init__(self):
        self.key: Optional[bytes] = None
        self.nonce: Optional[bytes] = None
        if self.encryption_type == "OLD_AES":
            if len(self.key_id) > 0:
                logger.debug("Attempting to create decryption key from DASH manifest")
                try:
                    self.key, self.nonce = decrypt_manifest_key_id(self.key_id)
                except Exception as e:
                    logger.exception(e)


@dataclass
class XMLDASHManifest:
    mime_type: Optional[str] = field(default=None)
    codecs: Optional[str] = field(default=None)
    content_type: Optional[str] = field(default=None)
    bandwidth: Optional[str] = field(default=None)
    audio_sampling_rate: Optional[str] = field(default=None)
    timescale: Optional[str] = field(default=None)
    initialization: Optional[str] = field(default=None, repr=False)
    media: Optional[str] = field(default=None, repr=False)
    start_number: Optional[str] = field(default=None, repr=False)
    segment_timeline: Optional["SegmentTimeline"] = field(default=None, repr=False)

    def __post_init__(self):
        # Initialize key and nonce even though they won't be used in
        # this manifest class
        self.key, self.nonce = None, None

        self.bandwidth: Optional[int] = (
            int(self.bandwidth) if self.bandwidth is not None else None
        )
        self.audio_sampling_rate: Optional[int] = (
            int(self.audio_sampling_rate)
            if self.audio_sampling_rate is not None
            else None
        )
        self.timescale: Optional[int] = (
            int(self.timescale) if self.timescale is not None else None
        )
        self.startNumber: Optional[int] = (
            int(self.start_number) if self.start_number is not None else None
        )

    def build_urls(self, session: Session) -> Optional[List[str]]:
        """Parse the MPEG-DASH manifest into a list of URLs. In
        particular, look for a special value, r, in self.segment_timeline.s.
        If there is no such value, set r=1. In both cases, start substituting
        r into the special substring, '$Number$', in self.initialization.
        Continue incrementing r and substituting until the resulting string
        returns a 500 error to a HEAD request."""
        if len(self.segment_timeline.s) == 0:
            return

        def sub_number(n: int, p: str = r"\$Number\$", s: str = self.media) -> str:
            return re.sub(p, str(n), s)

        try:
            r: Optional[int] = next(S.r for S in self.segment_timeline.s)
        except StopIteration:
            r = None

        # New path for when r is None; e.g. TIDAL track 96154223
        if r is None:
            urls_list: List[str] = [self.initialization]
            number: int = 1
            while session.head(url=sub_number(number)).status_code != 500:
                urls_list.append(sub_number(number))
                number += 1
            else:
                return urls_list
        else:
            number_range = range(self.startNumber, r + 1)  # include value of `r`
            urls_list: List[str] = [self.initialization] + [
                sub_number(i) for i in number_range
            ]
            number: int = r + 1
            while session.head(url=sub_number(number)).status_code != 500:
                urls_list.append(sub_number(number))
                number += 1
            else:
                return urls_list


Manifest = Union[JSONDASHManifest, XMLDASHManifest]


def manifester(tesrj: TracksEndpointStreamResponseJSON) -> Manifest:
    """Attempt to return a Manifest-type object based on
    the attributes of `tesrj`. Will raise TidalManifestException upon
    error"""
    if tesrj.manifest_mime_type == "application/vnd.tidal.bts":
        if tesrj.audio_mode not in {"DOLBY_ATMOS", "SONY_360RA", "STEREO"}:
            raise TidalManifestException(
                "Expected a manifest of Dolby Atmos, MQA, Sony 360 Reality Audio, "
                f"or encrypted-for-Windows-client audio for track {tesrj.track_id}"
            )

        try:
            manifest: Manifest = dataclass_wizard.fromdict(
                JSONDASHManifest, json.loads(tesrj.manifest_bytes)
            )
        except json.decoder.JSONDecodeError:
            raise TidalManifestException(
                f"Cannot parse manifest with type '{tesrj.manifest_mime_type}' as JSON"
            )
        except dataclass_wizard.errors.ParseError as pe:
            raise TidalManifestException(pe.message.split("\n")[0])

        if manifest.encryption_type == "NONE":
            return manifest
        elif manifest.encryption_type == "OLD_AES":
            if (manifest.key is not None) and (manifest.nonce is not None):
                return manifest
            else:
                raise TidalManifestException(
                    f"Audio data for track {tesrj.track_id}, audio mode "
                    f"{tesrj.audio_mode} could not be decrypted"
                )
        else:
            raise TidalManifestException(
                f"Audio data for track {tesrj.track_id}, audio mode "
                f"{tesrj.audio_mode} is incorrigibly encrypted with "
                f"encryption type '{manifest.encryption_type}'"
            )
    elif tesrj.manifest_mime_type == "application/dash+xml":
        try:
            xml: ET.Element = ET.fromstring(tesrj.manifest_bytes)
        except ET.ParseError:
            raise TidalManifestException(
                f"Expected an XML manifest for track {tesrj.track_id}"
            )

        ns: str = re.match(r"({.*})", xml.tag).groups()[0]
        st: SegmentTimeline = SegmentTimeline(
            tuple(
                S(**el.attrib) if el is not None else None
                for el in xml.findall(f".//{ns}S")
            )
        )

        manifest = XMLDASHManifest(
            xml.find(f".//{ns}AdaptationSet").get("mimeType"),
            xml.find(f".//{ns}Representation").get("codecs"),
            xml.find(f".//{ns}AdaptationSet").get("contentType"),
            xml.find(f".//{ns}Representation").get("bandwidth"),
            xml.find(f".//{ns}Representation").get("audioSamplingRate"),
            xml.find(f".//{ns}SegmentTemplate").get("timescale"),
            xml.find(f".//{ns}SegmentTemplate").get("initialization"),
            xml.find(f".//{ns}SegmentTemplate").get("media"),
            xml.find(f".//{ns}SegmentTemplate").get("startNumber"),
            st,
        )
        return manifest



from dataclasses import dataclass
import json
import logging
import math
from pathlib import Path
import shutil
import sys
from types import SimpleNamespace
from typing import Dict, List, Optional, Set, Tuple, Union

import ffmpeg
import mutagen
from requests import HTTPError, Session

from .media import AudioFormat
from .models import (
    PlaylistsEndpointResponseJSON,
    TracksEndpointResponseJSON,
    VideosEndpointResponseJSON,
)
from .requesting import request_playlists
from .track import Track
from .utils import (
    download_cover_image,
    replace_illegal_characters,
    temporary_file,
    TIDAL_API_URL,
)
from .video import Video

logger = logging.getLogger("__name__")


@dataclass
class Playlist:
    playlist_id: str  # UUID4

    def __post_init__(self):
        self.playlist_dir: Optional[Path] = None
        self.playlist_cover_saved: bool = False

    def set_metadata(self, session: Session):
        """Request from TIDAL API /playlists endpoint"""
        self.metadata: Optional[PlaylistsEndpointResponseJSON] = request_playlists(
            session=session, playlist_id=self.playlist_id
        )

        if self.metadata is None:
            return

        self.name = replace_illegal_characters(self.metadata.title)

    def set_items(self, session: Session):
        """Uses data from TIDAL API /playlists/items endpoint to
        populate self.items"""
        playlist_items: Optional[PlaylistsItemsResponseJSON] = get_playlist(
            session=session, playlist_id=self.playlist_id
        )
        if playlist_items is None:
            self.items = tuple()
        else:
            # For now, if playlist size, N, exceeds 100 items,
            # N items are returned, and all those between 101 and N
            # are simply None. It's a temporary fix before properly
            # handling paginated API requests, but it should stop
            # NoneType errors!
            self.items: Tuple[Optional[PlaylistItem]] = tuple(
                filter(None, playlist_items.items)
            )

    def set_playlist_dir(self, out_dir: Path):
        """Populates self.playlist_dir based on self.name, self.playlist_id"""
        playlist_substring: str = f"{self.name} [{self.playlist_id}]"
        self.playlist_dir: Path = out_dir / "Playlists" / playlist_substring
        self.playlist_dir.mkdir(parents=True, exist_ok=True)

    def save_cover_image(self, session: Session, out_dir: Path):
        """Requests self.metadata.image and attempts to write it to disk"""
        if self.playlist_dir is None:
            self.set_playlist_dir(out_dir=out_dir)
        self.cover_path: Path = self.playlist_dir / "cover.jpg"
        if not self.cover_path.exists():
            download_cover_image(
                session=session,
                cover_uuid=self.metadata.square_image,
                output_dir=self.playlist_dir,
                dimension=1080,
            )
        else:
            self.playlist_cover_saved = True

    def save_description(self):
        """Requests self.metadata.description and attempts to write it to disk"""
        self.description_path: Path = self.playlist_dir / "PlaylistDescription.txt"
        if self.metadata.description is not None and len(self.metadata.description) > 0:
            if not self.description_path.exists():
                self.description_path.write_text(f"{self.metadata.description}\n")

    def get_items(
        self, session: Session, audio_format: AudioFormat, no_extra_files: bool
    ):
        """Using either Track.get() or Video.get(), attempt to request
        the data for each track or video in self.items. If no_extra_files
        is True, do not attempt to retrieve or save any of: playlist
        description text, playlist m3u8 text, playlist cover image."""
        if len(self.items) == 0:
            return
        tracks_videos: list = [None] * len(self.items)
        for i, item in enumerate(self.items):
            if item is None:
                tracks_videos[i] = None
                continue
            elif isinstance(item, TracksEndpointResponseJSON):
                track: Track = Track(track_id=item.id)
                track.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=self.playlist_dir,
                    metadata=item,
                    no_extra_files=no_extra_files,
                )
                tracks_videos[i] = track
            elif isinstance(item, VideosEndpointResponseJSON):
                video: Video = Video(video_id=item.id)
                video.get(
                    session=session,
                    out_dir=self.playlist_dir,
                    metadata=item,
                    no_extra_files=no_extra_files,
                )
                tracks_videos[i] = video
            else:
                tracks_videos[i] = None
                continue
        else:
            self.tracks_videos: Tuple[Tuple[int, Optional[Union[Track, Video]]]] = (
                tuple(tracks_videos)
            )
        return tracks_videos

    def flatten_playlist_dir(self):
        """When self.get_items() is called, the tracks and/or videos in
        self.items are downloaded using their self-contained .get() logic;
        this means that they will be downloaded to albums. This function
        "flattens" self.playlist_dir, meaning that it moves all downloaded
        audio and video files to self.playlist_dir, and removes the various
        subdirectories created"""
        files: List[Dict[int, Optional[str]]] = [None] * len(self.tracks_videos)
        if len(self.tracks_videos) == 0:
            return
        subdirs: Set[Path] = set()

        for i, tv in enumerate(self.tracks_videos, 1):
            if getattr(tv, "outfile") is None:
                try:
                    getattr(tv, "album_dir")
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
                files[i - 1] = {i: None}
                continue

            _path: Optional[Path] = Path(tv.outfile) if tv is not None else None
            # if the item never got turned into a track or video
            if _path is None:
                files[i - 1] = {i: None}
                continue

            # if the track or video didn't download
            if _path.exists():
                if _path.stat().st_size == 0:
                    files[i - 1] = {i: None}
                    continue
            else:
                files[i - 1] = {i: None}
                continue

            # otherwise, move files and clean up
            if isinstance(tv, Track):
                new_path: Path = self.playlist_dir / f"{i:03d} - {tv.trackname}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
            elif isinstance(tv, Video):
                new_path: Path = self.playlist_dir / f"{i:03d} - {_path.name}"
                new_path.write_bytes(_path.read_bytes())
                _path.unlink()
                files[i - 1] = {i: str(new_path.absolute())}
        else:
            self.files: List[Dict[int, Optional[str]]] = files

        # Find all subdirectories written to
        subdirs: Set[Path] = set()
        for tv in self.tracks_videos:
            if isinstance(tv, Track):
                try:
                    getattr(tv, "album_dir")
                except AttributeError:
                    pass
                else:
                    subdirs.add(tv.album_dir)
                    subdirs.add(tv.album_dir.parent)
            elif isinstance(tv, Video):
                subdirs.add(tv.artist_dir)

        # Copy all artist images, artist bio JSON files out
        # of subdirs
        artist_images: Set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*.jpg"):
                if p.name == "cover.jpg":
                    continue
                artist_images.add(p)
        else:
            for artist_image_path in artist_images:
                if artist_image_path.exists():
                    shutil.copyfile(
                        artist_image_path.absolute(),
                        self.playlist_dir / artist_image_path.name,
                    )

        artist_bios: Set[Path] = set()
        for subdir in subdirs:
            for p in subdir.glob("*bio.json"):
                artist_bios.add(p)
        else:
            for artist_bio_path in artist_bios:
                if artist_bio_path.exists():
                    shutil.copyfile(
                        artist_bio_path.absolute(),
                        self.playlist_dir / artist_bio_path.name,
                    )

        # Remove all subdirs
        for subdir in subdirs:
            if subdir.exists():
                shutil.rmtree(subdir)
        else:
            return self.playlist_dir

    def craft_m3u8_text(self):
        """This method creates a file called playlist.m3u8 in self.playlist_dir
        that is a standard M3U. Needs to be called after self.flatten_playlist_dir
        in order to be able to access self.files
        N.b. the already-written file is temporarily copied to a .mp4 version in a
        temporary directory because .m4a files cannot be read with mutagen."""
        m3u_text: str = (
            f"#EXTM3U\n#EXTENC:UTF-8\n#EXTIMG:{str(self.cover_path.absolute())}\n#PLAYLIST:{self.name}\n"
        )

        logger.info(
            f"Creating .m3u8 playlist file for Playlist with ID '{self.playlist_id}'"
        )
        for d in self.files:
            file: str = next(iter(d.values()))
            if file is None:
                continue
            elif file.endswith(".flac"):
                m = mutagen.File(file)
                artist: str = m.get("artist", [""])[0]
                title: str = m.get("title", [""])[0]
                extinf: str = (
                    f"#EXTINF:{math.ceil(m.info.length)},"
                    f"{artist} - {title}\n{file}\n"
                )
                m3u_text += extinf
            elif file.endswith(".mka"):
                m = mutagen.File(file)
                artist: str = m.get("ARTI", [""])[0]
                title: str = m.get("TITL", [""])[0]
                extinf: str = (
                    f"#EXTINF:{math.ceil(m.info.length)},"
                    f"{artist} - {title}\n{file}\n"
                )
                m3u_text += extinf
            elif file.endswith(".m4a"):
                # Mutagen cannot read .m4a files, so make a copy with all
                # of the metadata tags as a .mp4 in a temporary directory
                with temporary_file(suffix=".mp4") as tf:
                    ffmpeg.input(file, hide_banner=None, y=None).output(
                        tf.name,
                        acodec="copy",
                        vcodec="copy",
                        loglevel="quiet",
                    ).run()

                    m = mutagen.File(tf.name)
                    artist: str = m.get("\xa9ART", [""])[0]
                    title: str = m.get("\xa9nam", [""])[0]
                    extinf: str = (
                        f"#EXTINF:{math.ceil(m.info.length)},"
                        f"{artist} - {title}\n{file}\n"
                    )
                    m3u_text += extinf
        else:
            return m3u_text

    def dumps(self):
        """This method emulates the stdlib json.dumps(). In particular,
        it returns the JSON-formatted string of the self.files attribute,
        which is an array of objects with one key each: the index in the
        playlist of the track or video, and the absolute file path to the
        track or video"""
        return json.dumps(self.files)

    def dump(self, fp=sys.stdout):
        """This method emulates the stdlib json.dump(). In particular,
        it sends to 'fp' the JSON-formatted string of the self.files attribute,
        which is an array of objects with one key each: the index in the
        playlist of the track or video, and the absolute file path to the
        track or video"""
        json.dump(self.files, fp)

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        no_extra_files: bool,
    ):
        """The main method of this class, executing a number of other methods
        in a row:
          - self.set_metadata()
          - self.set_items()
          - self.set_playlist_dir()
          - self.get_items()
          - self.flatten_playlist_dir()
        Then, if no_extra_files is False,
          - self.save_cover_image()
          - self.save_description()
          - self.craft_m3u8_text()
        """
        self.set_metadata(session)

        if self.metadata is None:
            self.files = {}
            return

        self.set_items(session)
        self.set_playlist_dir(out_dir)

        if self.get_items(session, audio_format, no_extra_files) is None:
            logger.critical(f"Could not retrieve playlist with ID '{self.playlist_id}'")
            self.files = {}
            return

        self.flatten_playlist_dir()
        logger.info(f"Playlist files written to '{self.playlist_dir}'")

        if not no_extra_files:
            try:
                self.save_description()
            except Exception:
                pass
            else:
                logger.info(
                    "Playlist description written to "
                    f"{self.playlist_dir / 'PlaylistDescription.txt'}"
                )

            self.save_cover_image(session, out_dir)

            try:
                m3u8_text: str = self.craft_m3u8_text()
            except Exception as e:
                logger.warning(
                    "Unable to create playlist.m3u8 file for "
                    f"playlist with ID '{self.playlist_id}'"
                )
                logger.debug(e)
            else:
                (self.playlist_dir / "playlist.m3u8").write_text(m3u8_text)
                logger.info(
                    "Playlist M3U file written to "
                    f"{self.playlist_dir / 'playlist.m3u8'}"
                )

    def get_elements(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        no_extra_files: bool,
    ):
        """The main method of this class when no_flatten is True at
        the program top level. It executes a number of other methods
        in a row:
          - self.set_metadata()
          - self.set_items()
        """
        self.set_metadata(session)

        if self.metadata is None:
            self.files = {}
            return

        self.set_items(session)
        if len(self.items) == 0:
            self.files = {}
            return
        else:
            files: List[Dict[int, Optional[str]]] = [None] * len(self.items)

        for i, item in enumerate(self.items):
            if item is None:
                files[i] = {i: None}
                continue
            elif isinstance(item, TracksEndpointResponseJSON):
                track: Track = Track(track_id=item.id)
                track_file: Optional[str] = track.get(
                    session=session,
                    audio_format=audio_format,
                    out_dir=out_dir,
                    metadata=item,
                    no_extra_files=no_extra_files,
                )
                files[i] = {i: track_file}
            elif isinstance(item, VideosEndpointResponseJSON):
                video: Video = Video(video_id=item.id)
                video_file: Optional[str] = video.get(
                    session=session,
                    out_dir=out_dir,
                    metadata=item,
                )
                files[i] = {i: video_file}
            else:
                files[i] = {i: None}
                continue
        else:
            self.files: List[Dict[int, Optional[str]]] = files


class TidalPlaylistException(Exception):
    pass


def request_playlist_items(session: Session, playlist_id: str) -> Optional[dict]:
    """Request from TIDAL API /playlists/items endpoint. If requests.HTTPError
    arises, warning is logged; upon this or any other exception, None is returned.
    If no exception arises from 'session'.get(), the requests.Response.json()
    object is returned."""
    url: str = f"{TIDAL_API_URL}/playlists/{playlist_id}/items"
    kwargs: dict = {"url": url}
    kwargs["params"] = {"limit": 100}
    kwargs["headers"] = {"Accept": "application/json"}

    data: Optional[dict] = None
    logger.info(f"Requesting from TIDAL API: playlists/{playlist_id}/items")
    with session.get(**kwargs) as resp:
        try:
            resp.raise_for_status()
        except HTTPError as he:
            if resp.status_code == 404:
                logger.warning(
                    f"404 Client Error: not found for TIDAL API endpoint playlists/{playlist_id}/items"
                )
            else:
                logger.exception(he)
        else:
            data = resp.json()
            logger.debug(
                f"{resp.status_code} response from TIDAL API to request: playlists/{playlist_id}/items"
            )
        finally:
            return data


@dataclass(frozen=True)
class PlaylistsItemsResponseJSON:
    """The response from the TIDAL API endpoint /playlists/<ID>/items
    is modeled by this class."""

    limit: int
    offset: int
    total_number_of_items: int
    items: Tuple[
        Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]
    ]


def playlist_maker(
    playlists_response: Dict[str, Union[int, List[dict]]]
) -> "PlaylistsItemsResponseJSON":
    """This function massages the response from the TIDAL API endpoint
    playlists/items into a format that PlaylistsItemsResponseJSON.from_dict()
    can ingest. Returns a PlaylistsItemsResponseJSON instance"""
    init_args: dict = {}
    init_args["limit"] = playlists_response.get("limit")
    init_args["offset"] = playlists_response.get("offset")
    init_args["total_number_of_items"] = playlists_response.get("totalNumberOfItems")

    items: Tuple[SimpleNamespace] = tuple(
        SimpleNamespace(**d) for d in playlists_response["items"]
    )
    if len(items) == 0:
        return

    playlist_items: List[
        Optional[Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]]
    ] = [None] * init_args["total_number_of_items"]

    for i, namespace in enumerate(items):
        if namespace.type == "track":
            try:
                playlist_item = TracksEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                logger.warning(
                    f"TidalPlaylistException: unable to parse playlist item [i] "
                    f"with type '{namespace.type}'"
                )
                logger.debug(e)
                # value stays None
            else:
                playlist_items[i] = playlist_item
        elif namespace.type == "video":
            try:
                playlist_item = VideosEndpointResponseJSON.from_dict(namespace.item)
            except Exception as e:
                logger.warning(
                    f"TidalPlaylistException: unable to parse playlist item [i] "
                    f"with type '{namespace.type}'"
                )
                logger.debug(e)
                # value stays None
            else:
                playlist_items[i] = playlist_item
        else:
            continue  # value stays None
    else:
        init_args["items"] = tuple(playlist_items)

    return PlaylistsItemsResponseJSON(**init_args)


def get_playlist(
    session: Session, playlist_id: str
) -> Optional["PlaylistsItemsResponseJSON"]:
    """The pattern for playlist items retrieval does not follow the
    requesting.request_* functions, hence its implementation here.
    """
    playlists_items_response_json: Optional["PlaylistsItemsResponseJSON"] = None
    try:
        playlists_response: dict = request_playlist_items(
            session=session, playlist_id=playlist_id
        )
        playlists_items_response_json: Optional["PlaylistsItemsResponseJSON"] = (
            playlist_maker(playlists_response=playlists_response)
        )
    except Exception as e:
        logger.exception(TidalPlaylistException(e.args[0]))
    finally:
        return playlists_items_response_json


# union type for type hinting
PlaylistItem = Optional[
    Union["TracksEndpointResponseJSON", "VideosEndpointResponseJSON"]
]



from dataclasses import dataclass
import json
import logging
from pathlib import Path
import sys
from typing import List, Optional, Tuple

from requests import Session

from .media import AudioFormat
from .models import (
    AlbumsCreditsResponseJSON,
    AlbumsEndpointResponseJSON,
    AlbumsItemsResponseJSON,
    AlbumsReviewResponseJSON,
    TracksEndpointResponseJSON,
)
from .requesting import (
    request_albums,
    request_albums_credits,
    request_album_items,
    request_album_review,
)
from .track import Track
from .utils import download_cover_image

logger = logging.getLogger("__name__")


@dataclass
class Album:
    album_id: int

    def __post_init__(self):
        self.album_dir: Optional[Path] = None
        self.album_cover_saved: bool = False

    def set_tracks(self, session: Session):
        """This method populates self.tracks by requesting from
        TIDAL albums/items endpoint."""
        album_items: AlbumsItemsResponseJSON = request_album_items(
            session=session, album_id=self.album_id
        )
        _items = album_items.items if album_items is not None else ()
        self.tracks: Tuple[TracksEndpointResponseJSON] = tuple(
            _item.item for _item in _items
        )

    def set_metadata(self, session: Session):
        """This method sets self.metadata by requesting from
        TIDAL /albums endpoint"""
        self.metadata: AlbumsEndpointResponseJSON = request_albums(
            session=session, album_id=self.album_id
        )

    def set_album_review(self, session: Session):
        """This method requests the review corresponding to self.album_id
        in TIDAL. If it exists, it is written to disk as AlbumReview.json
        in self.album_dir"""
        self.album_review: Optional[AlbumsReviewResponseJSON] = request_album_review(
            session=session, album_id=self.album_id
        )
        if self.album_review is not None:
            (self.album_dir / "AlbumReview.json").write_text(
                self.album_review.to_json()
            )

    def set_album_credits(self, session: Session):
        """This method requests the album's top-level credits (separate from
        each track's credits) and writes them to AlbumCredits.json in
        self.album_dir"""
        self.album_credits: Optional[AlbumsCreditsResponseJSON] = (
            request_albums_credits(session=session, album_id=self.album_id)
        )
        if self.album_credits is not None:
            f: str = str((self.album_dir / "AlbumCredits.json").absolute())
            with open(f, "w") as fp:
                json.dump(obj=self.album_credits.credit, fp=fp)

    def set_album_dir(self, out_dir: Path):
        """This method populates self.album_dir as a sub-subdirectory of
        out_dir: its parent directory is the name of the (main) artist of
        the album"""
        artist_substring: str = self.metadata.artist.name.replace("..", "")
        album_substring: str = (
            f"{self.metadata.name.replace('..', '')} "
            f"[{self.metadata.id}] [{self.metadata.release_date.year}]"
        )
        self.album_dir = out_dir / artist_substring / album_substring
        self.album_dir.mkdir(parents=True, exist_ok=True)

        if self.metadata.number_of_volumes > 1:
            for v in range(1, self.metadata.number_of_volumes + 1):
                volume_substring: str = f"Volume {v}"
                (out_dir / artist_substring / album_substring / volume_substring).mkdir(
                    parents=True, exist_ok=True
                )

    def save_cover_image(self, session: Session, out_dir: Path):
        """This method writes cover.jpg in self.album_dir via the
        utils.download_cover_image() function. If successful,
        then self.album_cover_saved takes the value True"""
        if self.album_dir is None:
            self.set_album_dir(out_dir=out_dir)
        self.cover_path: Path = self.album_dir / "cover.jpg"
        if not self.cover_path.exists():
            download_cover_image(
                session=session,
                cover_uuid=self.metadata.cover,
                output_dir=self.album_dir,
            )
        else:
            self.album_cover_saved = True

    def get_tracks(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        no_extra_files: bool,
    ) -> List[Optional[str]]:
        """This method uses self.tracks to call track.Track.get() for each
        element in self.tracks. It uses the result of each of these calls to
        populate self.track_files"""
        track_files: List[str] = [None] * self.metadata.number_of_tracks
        for i, t in enumerate(self.tracks):  # type(t) is TracksEndpointResponseJSON
            track: Track = Track(track_id=t.id)

            track_files_value: Optional[str] = track.get(
                session=session,
                audio_format=audio_format,
                out_dir=out_dir,
                metadata=t,
                album=self.metadata,
                no_extra_files=no_extra_files,
            )
            track_files[i] = {track.metadata.track_number: track_files_value}
        else:
            self.track_files = track_files

    def dumps(self):
        """This method returns a JSON-like string of self.track_files"""
        return json.dumps(self.track_files)

    def dump(self, fp=sys.stdout):
        """This method writes to (by default) STDOUT a
        JSON-like string of self.track_files"""
        json.dump(self.track_files, fp)

    def get(
        self,
        session: Session,
        audio_format: AudioFormat,
        out_dir: Path,
        metadata: Optional[AlbumsEndpointResponseJSON] = None,
        no_extra_files: bool = False,
    ):
        """This method is the driver method of the class. It calls the
        other methods in order:
            1. set_metadata()
            2. set_tracks()
            3. save_cover_image()
            4. set_album_review()
            5. get_tracks()
        """
        if metadata is None:
            self.set_metadata(session)
        else:
            self.metadata = metadata

        if self.metadata is None:
            self.track_files = {}
            return

        self.set_tracks(session)

        self.set_album_dir(out_dir)

        if self.metadata.cover != "":  # None was sent from the API
            self.save_cover_image(session, out_dir)
        else:
            logger.warning(
                f"No cover image was returned from TIDAL API for album {self.album_id}"
            )

        if not no_extra_files:
            self.set_album_review(session)
            self.set_album_credits(session)
        else:
            try:
                self.cover_path.unlink()
            except FileNotFoundError:
                pass

        self.get_tracks(session, audio_format, out_dir, no_extra_files)
